#include <iostream>
#include <vector>
#include <fstream>
#include <string>
#include <iomanip>
#include <ctime>

class Terrain
{
public :
	struct InitInfo
	{
		std::wstring HeightMapFilename ;
		float		 HeightScale ;
		unsigned int HeightmapWidth ;
		unsigned int HeightmapHeight ;
	} ;

public :
	Terrain( Terrain::InitInfo initInfo ) ;
	~Terrain() ;

	void Smooth() ;

private :
	float Average( int i, int j ) ;
	bool  InBounds( int i, int j ) ;

private :
	std::vector< float >	mHeightmap ;
	InitInfo				mInfo ;
} ;

Terrain::Terrain( Terrain::InitInfo initInfo ) : mInfo( initInfo )
{
	// A height for each vertex
	std::vector< unsigned char > in( mInfo.HeightmapWidth * mInfo.HeightmapHeight ) ;

	// Open the file.
	std::ifstream inFile ;
	inFile.open( mInfo.HeightMapFilename.c_str(), std::ios_base::binary ) ;

	if( inFile )
	{
		// Read the RAW bytes.
		//inFile.read( ( char* )&in[ 0 ], ( std::streamsize )in.size() ) ;
		inFile.read( reinterpret_cast< char* >( &in[ 0 ] ), static_cast< std::streamsize >( in.size() ) ) ;

		// Done with file.
		inFile.close() ;
	}
	else
	{
		std::cerr << "File could not be opened!" << std::endl ;
		system( "pause" ) ;
		exit( 1 ) ;
	}

	// Copy the array data into a float array and scale it.
	mHeightmap.resize( mInfo.HeightmapHeight * mInfo.HeightmapWidth, 0 ) ;
	for( unsigned int i = 0 ; i < mInfo.HeightmapHeight * mInfo.HeightmapWidth ; ++i )
	{
		mHeightmap[ i ] = ( in[ i ] / 255.0f ) * mInfo.HeightScale ;
	}
}

Terrain::~Terrain()
{}

void Terrain::Smooth()
{
	std::vector< float > dest( mHeightmap.size() ) ;

	for( unsigned int i = 0 ; i < mInfo.HeightmapHeight ; ++i )
	{
		for( unsigned int j = 0 ; j < mInfo.HeightmapWidth ; ++j )
		{
			dest[ i * mInfo.HeightmapWidth + j ] = Average( i, j ) ;
		}
	}

	// Replace the old heightmap with the filtered one.
	mHeightmap = dest ;
}

bool Terrain::InBounds( int i, int j )
{
	// True if ij are valid indices; false otherwise.
	return i >= 0 && i < static_cast< int >( mInfo.HeightmapHeight ) &&
		   j >= 0 && j < static_cast< int >( mInfo.HeightmapWidth ) ;
}

float Terrain::Average( int i, int j )
{
	// Function computes the average height of the ij element.
	// It averages itself with its eight neighbor pixels.  Note
	// that if a pixel is missing neighbor, we just don't include it
	// in the average--that is, edge pixels don't have a neighbor pixel.
	//
	// ----------
	// | 1| 2| 3|
	// ----------
	// |4 |ij| 6|
	// ----------
	// | 7| 8| 9|
	// ----------

	float avg = 0.0f ;
	float num = 0.0f ;

	// Use int to allow negatives.  If we use UINT, @ i=0, m=i-1=UINT_MAX
	// and no iterations of the outer for loop occur.
	for( int m = i - 1 ; m <= i + 1 ; ++m )
	{
		for( int n = j - 1 ; n <= j + 1 ; ++n )
		{
			if( InBounds( m, n ) )
			{
				avg += mHeightmap[ m * mInfo.HeightmapWidth + n ] ;
				num += 1.0f ;
			}
		}
	}

	return avg / num ;
}

int main()
{
	// Fill InitInfo.
	Terrain::InitInfo terrainInitInfo  ;
	terrainInitInfo.HeightMapFilename = L"terrain.raw"  ;
	terrainInitInfo.HeightScale		  = 50.0f  ;
	terrainInitInfo.HeightmapWidth	  = 2049  ;
	terrainInitInfo.HeightmapHeight   = 2049  ;

	// 2 digits after dot.
	std::cout << std::fixed << std::setprecision( 2 ) ; 

	// Load heightmap.
	std::cout << "Loading heightmap...   " ;

	clock_t t1 = clock() ;
	Terrain terrain( terrainInitInfo ) ;
	clock_t t2 = clock() ;

	double t2Minust1 = double( t2 - t1 ) / CLOCKS_PER_SEC ;

	std::cout << "time elapsed = " << t2Minust1 << " seconds!" << std::endl ;

	// Smooth heightmap.
	std::cout << "Smoothing...           " ;

	t1 = clock() ;
	terrain.Smooth() ;
	t2 = clock() ;

	t2Minust1 = double( t2 - t1 ) / CLOCKS_PER_SEC ;

	std::cout << "time elapsed = " << t2Minust1 << " seconds!" << std::endl ;

	// Done.
	std::cout << "\a" ;
	system( "pause" ) ;
	return 0 ;
}